#include "legato.h"
#include "interfaces.h"
#include "le_data_interface.h"
//#include "le_info_interface.h"
#include "le_timer.h"
#include "le_print.h"
#include "errno.h"
#include "le_clock.h"
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netdb.h>
#include <stdlib.h>
//#include <std>
#include <string.h>
#include <string>
using namespace std;
#include "json/json.h"
// Needs to be last include
#include "CRCPP.hpp"


//extern int h_errno;

static int32_t crReportingInterval;
static le_sem_Ref_t    crSemaphore;
static le_timer_Ref_t crTimer;
static le_clk_Time_t crTimerStruct;


static CRRPTR * pCRRPTR;

CRRPTR::CRRPTR()
{
		crReportingInterval = 2;
		crURL = "";
		crPort = 0;
		crReportingEnabled = false;
		crSemaphore = NULL;
		crTimer = NULL;
		firmware = "";
		mFirmware = "";
		carrier = "";
		IMEI = "";
		reportBuffer = "";
	}

void CRRPTR::ReadCRValues(void)
{
   //ToDo  Add file read of the set values.  For now, these will the be defaults

   crURL = "xrdl001.feeneywireless.com";

   crReportingInterval = 2;
   crPort = 39759;
   crReportingEnabled = true;
}
void CRRPTR::GetCarrier()
{
   char tmpStr[100];
   le_sim_Id_t sim = le_sim_GetSelectedCard();
   if (le_sim_GetHomeNetworkOperator(sim, tmpStr, sizeof(tmpStr)) == LE_OK)
   {
      carrier = tmpStr;
   }
   if (le_sim_GetICCID(sim, tmpStr, sizeof(tmpStr)) == LE_OK)
   {
      iccid = tmpStr;
   }

}
void CRRPTR::GetIMEI()
{
	char tmpStr[50];
	le_info_GetImei(tmpStr,sizeof(tmpStr));
	IMEI = tmpStr;
}

string CRRPTR::GetSQLFormatDateTime()
{
    time_t t1;
    struct tm tms;
    time(&t1);
    localtime_r(&t1, &tms);
    char tmp[255];

    sprintf(tmp, "%d-%0d-%0d %0d:%0d:%0d",tms.tm_year+1900, tms.tm_mon+1,tms.tm_mday, tms.tm_hour, tms.tm_min, tms.tm_sec);
    timeStamp = tmp;

    //string date = string.Format("{0:yyyy}-{0:MM}-{0:dd} {0:HH}:{0:mm}:{0:ss}",now);
    return(timeStamp);
}

void CRRPTR::GetSignalStateAndQuality(void)
{
    le_result_t           res;
    
 
    res = le_mrc_GetNetRegState(&regState);
    if (res == LE_OK)
      LE_INFO( "regState is: %d",regState);
   else 
      LE_INFO("GetNetRegState failed!");

    res = le_mrc_GetSignalQual(&signalQuality);

    
    //LE_ASSERT(quality<=5);
}


void CRRPTR::GetSignalMetrics(void)
{
    le_result_t   res;
    
    char buffer[100];

    le_mrc_MetricsRef_t metricsRef = le_mrc_MeasureSignalMetrics();
    LE_ASSERT(metricsRef != NULL);
 
    rat = le_mrc_GetRatOfSignalMetrics(metricsRef);
    LE_INFO("RAT of signal metrics is %d",rat);
    switch(rat)
    {
        case LE_MRC_RAT_GSM:
            res = le_mrc_GetGsmSignalMetrics(metricsRef, &rxLevel, &er);
            if (res == LE_OK)
            {
               LE_INFO("GSM metrics rxLevel.%ddBm, er.%d", rxLevel, er);
            }
            break;
 
        case LE_MRC_RAT_UMTS:
        case LE_MRC_RAT_TDSCDMA:
            res = le_mrc_GetUmtsSignalMetrics(metricsRef, &rxLevel, &er, &ecio, &rscp, &sinr);
            if(res == LE_OK)
            {
               sprintf(buffer, "UMTS/TD-SCDMA metrics rxLevel.%ddBm, er.%d, ecio.%010.1fdB,"
                    "rscp.%ddBm, sinr.%ddB", rxLevel, er, ((double)ecio/10), rscp, sinr);
               LE_INFO(buffer);
            }
            break;
 
        case LE_MRC_RAT_LTE:
            res = le_mrc_GetLteSignalMetrics(metricsRef, &rxLevel, &er, &rsrq, &rsrp, &snr);
            if(res == LE_OK)
            {

               LE_INFO("LTE metrics rxLevel.%ddBm, er.%d, rsrq.%010.1fdB, "
                    "rsrp.%010.1fdBm, snr.%010.1fdB",
                    rxLevel, er, ((double)rsrq/10), ((double)rsrp/10), ((double)snr/10));
            }
            break;
 
        case LE_MRC_RAT_CDMA:
            res = le_mrc_GetCdmaSignalMetrics(metricsRef,  &rxLevel, &er, &ecio, &sinr, &io);
            if (res == LE_OK)
            {
               LE_INFO("CDMA metrics rxLevel.%ddBm, er.%d, ecio.%010.1fdB, "
                    "sinr.%ddB, io.%ddBm",
                    rxLevel, er, ((double)ecio/10), sinr, io);
            }
            break;
 
        default:
            LE_FATAL("Unknown RAT!");
            break;
    }
 
    le_mrc_DeleteSignalMetrics(metricsRef);
}

void CRRPTR::DoSignalStuff(Json::Value &root)
{
   GetSignalStateAndQuality();
   GetSignalMetrics();
   switch(rat)
   {
      case LE_MRC_RAT_LTE:
         root["rsrq"] = rsrq/10;
         root["rsrp"] = rsrp/10;
         root["snr"] = snr/10;
         root["connType"] = "4GLTE";
      break;
      case LE_MRC_RAT_CDMA:
         root["rsrp"] = rxLevel;
         root["rsrq"] = ecio/10;
         root["snr"] = sinr;
         root["connType"] = "2G";
      break;
      case LE_MRC_RAT_UMTS:
      case LE_MRC_RAT_TDSCDMA:
         root["rssi"] = rxLevel;
         root["rsrq"] = ecio/10;
         root["snr"] = sinr;
         root["connType"] = "3G";
      break;
      default:
      break;
   }

}

void CRRPTR::BuildCRReportString()
{
	char tmpStr[100] = "";
	Json::Value root;
	//Json::Value child;
	Json::StreamWriterBuilder writer;
	string jsonString;
   string timestamp;

	le_mdc_ProfileRef_t profileRef  = NULL;
   writer["indentation"] = "";

	int32_t currentProfileIndex = le_data_GetCellularProfileIndex();

	profileRef = le_mdc_GetProfile(currentProfileIndex);
	

   

   GetCarrier();
	GetIMEI();
   DoSignalStuff(root);
  

	root["uid"] = IMEI;
	root["crinterval"] = crReportingInterval;
	root["deviceType"] = "Skyus-SC";
	root["carrier"] = carrier;

	root["imei"] = IMEI;
	root["timestamp"] = GetSQLFormatDateTime();

   le_info_GetDeviceModel(tmpStr, sizeof(tmpStr));
   root["deviceModel"]  = tmpStr;
   le_info_GetMdn(tmpStr, sizeof(tmpStr));
   root["ptn"] = tmpStr;
   le_info_GetMin(tmpStr, sizeof(tmpStr));
   root["min"] = tmpStr;
   le_info_GetFirmwareVersion (tmpStr, sizeof(tmpStr));
   root["firmware"] = tmpStr;
   root["mFirmware"] = ".01";  //To Do create a firmware version file for our app.

   le_mdc_GetIPv4Address(profileRef, tmpStr, sizeof(tmpStr));
   v4IPAddress = tmpStr;
   root["ipaddress"] = v4IPAddress;

   jsonString = Json::writeString(writer, root);

	//LE_INFO("jsonString is: %s",jsonString.c_str() );


   reportBuffer = jsonString;

   LE_PRINT_VALUE("%s",jsonString.c_str());


}

void CRRPTR::SendData(char * url, int portno)
{
	int sockfd, n;

	struct sockaddr_in serv_addr;
	struct hostent *server;
	char buffer[256];
	int i = 0;

	sockfd = socket(AF_INET, SOCK_STREAM,0);
	if (sockfd < 0)
	{
		LE_INFO("Can't open the socket to CR!");
		return;
	}
	server = gethostbyname((char *)url);
	while(server == NULL && i < 10)
	{
		i++;
		sleep(1);
		//retry
		server = gethostbyname((char *)url);
	}
	//Debug Fallback
	if (server == NULL)
	{
		server = gethostbyname((char *)"104.152.217.161");
		LE_INFO("Can't resolve CR URL %s, using debug fallback", url);
	}

	if (server == NULL)
	{
		LE_INFO("Can't resolve CR URL %s,  errno is %d", url, h_errno);
		close(sockfd);
		return;
	}

	bzero((char *) &serv_addr, sizeof(serv_addr));
	serv_addr.sin_family = AF_INET;
	bcopy((char*)server->h_addr,(char *)&serv_addr.sin_addr.s_addr,server->h_length);
	serv_addr.sin_port = htons(portno);
	if (connect(sockfd,(struct sockaddr *)&serv_addr,sizeof(serv_addr)) < 0)
	{
		LE_INFO("Can't connect to CR!");
		close(sockfd);
		return;
	}

	n = write(sockfd, reportBuffer.c_str(), reportBuffer.length()+1);
	if (n < 0)
	{
		LE_INFO("Can't write to CR Socket! errno is %d", h_errno);
		close(sockfd);
		return;
	}
	bzero(buffer,256);
	n = read(sockfd, buffer, 255);
	if (n < 0)
	{
		LE_INFO("CR - Nothing received!");
	}
	else if (buffer[0] != 0)
	{
		LE_INFO("CR returned %s",buffer);
	}
	close(sockfd);
}

void CRRPTR::CRReport()
{
	le_result_t res;
   le_data_RequestObjRef_t dataSession = NULL;
   int i = 0;

   // Wait for a data connection
   LE_INFO("Trying to report to Crossroads.");
   res = le_mrc_GetNetRegState(&regState);
   if (res == LE_OK)
   {
      if(regState != LE_MRC_REG_HOME  && regState != LE_MRC_REG_ROAMING)
      {  
         // No point in trying a data connection.
         LE_INFO("Modem not registered!  regState = %d", regState);
         return; //
      }
   }
   dataSession = le_data_Request();
   while (dataSession == NULL)
   {
      i++;
      if (i > 10)
      {
         LE_ERROR("CRReporting Couldn't establish a data connection, bombing out.");
         return;
      }
      sleep(10);
      dataSession = le_data_Request();
   }
   LE_INFO("CR - Got dataSession.");
   //sleep(5);  //Wait for DNS issues to resolve.
   //Build the JSON string.
   BuildCRReportString();

   LE_INFO(reportBuffer.c_str());

   //Send the data
   SendData((char *)crURL.c_str(), crPort);

   sleep(1);
   le_data_Release(dataSession);

}
void CRTimerHandler(le_timer_Ref_t crTimerRef)
{
   //Reset the timer value in case it changed
   crTimerStruct.sec = crReportingInterval*60;

   le_timer_SetInterval(  crTimer, crTimerStruct);

   le_timer_Start(crTimer);
   //Post the semaphore so the data can be sent.
   le_sem_Post(crSemaphore);
}

void DeleteTimerAndFreePtr(le_timer_Ref_t t)
{
    le_timer_Stop( t );
    free( le_timer_GetContextPtr( t ) );
    le_timer_Delete( t );  // timer ref is now invalid
}
void* TimerThread( void* contextPtr )
{
    //le_mdc_ConnectService();

    //le_mdc_ProfileRef_t profileRef = *((le_mdc_ProfileRef_t*)contextPtr);

    // Add state handler on the profile
    //le_mdc_AddSessionStateHandler(profileRef, StateChangeHandler, NULL);

   while (!crTimer)
   {
      crTimer = le_timer_Create("crTimer");
   }

   le_timer_SetHandler( crTimer, CRTimerHandler );

   le_timer_SetInterval(  crTimer, crTimerStruct);

   le_timer_Start(crTimer);

   // Post a semaphore to synchronize the timer
    le_sem_Post(crSemaphore);

    // Run the event loop
    LE_INFO("Run Event Loop...");
    le_event_RunLoop();
}


COMPONENT_INIT
{
	//sleep(10); uncomment out this if you have another connection manager.
	CRRPTR *CRReporter = new CRRPTR();

	CRReporter->Start();

	delete (CRReporter);
}

void CRRPTR::Start(void)
{
   LE_INFO("CRReporter Init");
   bool report = true;
   le_thread_Ref_t timerThread;
   le_result_t res;
   static le_clk_Time_t crTimeStruct;

   ReadCRValues();
   crTimeStruct.sec = crReportingInterval*60*2;
   crTimeStruct.usec = 0;


   crSemaphore = le_sem_Create("crSemaphore",0);

   timerThread = le_thread_Create("CRTimer", TimerThread, NULL);

     // Start a thread to treat the event handler.
    LE_INFO("Starting Event Handler thread.");
    le_thread_Start(timerThread);

    // Wait for the call of the event handler
    res = le_sem_WaitWithTimeOut(crSemaphore, crTimeStruct);

    if (res != LE_OK)
         LE_INFO("crSemaphore wait returned %d", res);

   //Top of loop
   while(report)
   {

      if (crReportingEnabled)
      {
         CRReport();
      }
      //crTimeStruct.sec = 2;


      res = le_sem_WaitWithTimeOut(crSemaphore, crTimeStruct);

      if (res != LE_OK)
         LE_INFO("crSemaphore wait returned %d", res);

      ReadCRValues(); //recheck each time through the loop in case something changes.


   }
   DeleteTimerAndFreePtr(crTimer);
   crTimer = NULL;
}

string CRRPTR::GetConnectSettings(void)
{
	string connectSettings;
	connectSettings = "minRSSI" + '=' + to_string(minRSSI) + ',';
	connectSettings += "minRSSI4G" + '=' + to_string(minRSSI4G) + ',';
	connectSettings += "pingInterval" + '=' + to_string(pingInterval);
	connectSettings += "pingURL" + '=' + pingURL + ',';
	connectSettings += "altPingURL" + '=' + altPingURL + ',';
	connectSettings += "pingRetries" + '=' + to_string(pingRetries) + ',';
	connectSettings += "pingRetryInterval" + '=' + to_string(pingRetryInterval);
	return (connectSettings);
}

void SServer_GetConnectionSettings(char* connSettings,size_t size)
{
   LE_INFO("GetConnectionSettings called");
   string empty = "";
   if (pCRRPTR != NULL)
     strncpy((char *)connSettings, pCRRPTR->GetConnectSettings().c_str(), size);
   else
     strncpy((char *)connSettings,empty.c_str(),size);

   LE_INFO("connection settings = %s", connSettings);

}
